#!/usr/bin/env python
#coding: utf-8

# sd_sapi.py - complete SAPI multi-language speech-dispatcher module
# Copyright (C) Bohdan R. Rau 2012 <ethanak@polip.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, write to:
# 	The Free Software Foundation, Inc.,
# 	51 Franklin Street, Fifth Floor
# 	Boston, MA  02110-1301, USA.


from spdsimple import SpeechModule

import sys,subprocess,os,os.path,re
import time

import sapilektor


class SapiSpeaker(SpeechModule):
    
    def __init__(self):
        self._phrasers={}
        self._sapi=sapilektor.Lektor()
        SpeechModule.__init__(self,"sapi",sys.argv)
        
    def _get_phraser(self,lang):
        if not lang in self._phrasers:
            if lang == 'de':
                from germanizer import Germanizer
                phraser=Germanizer(charset='utf-8')
            elif lang == 'pl':
                from polonizer import Polonizer
                phraser=Polonizer(charset='utf-8')
            else:
                from spdsimple.phonetizer import Latinizer
                phraser=Latinizer(charset='utf-8')
            self._phrasers[lang]=phraser
        return self._phrasers[lang]
            
    def init_module(self):
        """
        Overrides default class
        """
        self._phrasers={}
        self.sapi=sapilektor.Lektor()
        voices=self.sapi.GetVoiceList()
        self.voices={}
        self.voices_by_langmode={}
        self.defaultvoice=None
        for voice in voices:
            name=voice[0].lower()
            self.voices[name]={
                'gender':voice[2][0].lower(),
                'lang':voice[1],
                'phraser':self._get_phraser(voice[1])}
            if not self.defaultvoice:
                self.defaultvoice=name
            if voice[4] != 0:
                if not voice[1] in self.voices_by_langmode:
                    self.voices_by_langmode[voice[1]] = {}
                self.voices_by_langmode[voice[1]][voice[4]]=name
        if len(voices) == 0:
            self.log("No voices found")
            return False
        self.settings['synthesis_voice']=self.defaultvoice        
        self.log("Module init OK, DefVoice=%s" % self.defaultvoice)
        return True
        
    def enshort_message(self,message):
        om=[]
        icon = None
        mlen=int(self.settings.get('max_sentence_length',100))
        sm1=re.compile(r'^(.{5,%d}[;.,:]\s+)(.*)$' % (mlen-3),re.DOTALL | re.UNICODE)
        sm2=re.compile(r'^(.{5,%d})\s+(.*)$' % (mlen-3),re.DOTALL | re.UNICODE)
        
        for msg in message:
            if isinstance(msg,tuple):
                icon=msg[1]
                msg=msg[0]
            if not msg:
                if icon:
                    om.append((msg,icon))
                    icon=None
                continue
            if not isinstance(msg,unicode):
                msg=msg.decode('utf-8')
            msg=re.sub(r'\s+',' ',msg.strip())
            while len(msg) > mlen:
                r=sm1.match(msg)
                if not r:
                    r=sm2.match(msg)
                if r:
                    
                    m=r.group(1).strip().encode('utf-8')
                    if icon:
                        om.append(m,icon)
                        icon=None
                    else:
                        om.append(m)
                    msg=r.group(2)
                    continue
                break
            msg=msg.encode('utf-8')
            if icon:
                om.append((msg,icon))
                icon=None
            else:
                om.append(msg)
        return om
                
    def speaker(self):
        vname=self.settings.get('synthesis_voice',self.defaultvoice).lower()
        if not vname in self.voices:
            vname=self.defaultvoice
        voice=self.voices[vname]
        message=voice['phraser'].split(self.spd,self.command,self.settings)
        if not message:
            return
        if len(message) == 0:
            return
        if self.command == 'SPEAK':
            message=self.enshort_message(message)
        pitch=self.ensure(self.settings.get('pitch',0),-100,100)
        rate=self.ensure(self.settings.get('rate',0),-50,100)
        volume=self.ensure(self.settings.get('volume',0),-100,100)

        self.audio.vol(volume)
        self.audio.contrast(int(self.settings.get('contrast',0)))
        self.audio.setMargin(int(self.settings.get('pulse_audio_margin','300')))
        self.sapi.SetSynthesisVoice(self.get_current_voice())
        self.sapi.SetPauseFactor(self.settings['pause_factor'])

        while len(message)>0 and not self.audio.stopped:
            atime=time.time()
            msg=message.pop(0)
            icon=None
            if isinstance(msg,tuple):
                icon=msg[1]
                msg=msg[0]
            if not isinstance(msg,str):
                raise Exception("Message is not string!")
            self.log("Message %s" % msg)
            wave=self.sapi.GetWave(msg,True)
            if icon:
                self.audio.playFile(os.path.join(self.settings['sound_icons'],icon))
            if wave:
                self.audio.play(wave[0],freq=wave[1],rate=rate,pitch=pitch)
            self.audio.end()
    
    #overs
    
    def get_current_voice(self):
        v=self.settings.get('synthesis_voice',self.defaultvoice)
        if not v in self.voices:
            v=self.defaultvoice
        return v
    
    def get_voice(self,vv,va):
        v=self.settings.get('synthesis_voice',self.defaultvoice)
        if not v in self.voices:
            v=self.defaultvoice
        old_lang=self.voices[v]['lang']
        new_lang=self.settings['language']
        if 'voice' in self.settings:
            new_voice_type=self.settings['voice'].lower()
        else:
            new_voice_type='unknown'
        llist=self.voices_by_langmode.get(new_lang,old_lang)
        if not llist:
            return
        vid = self.get_voice_id(new_voice_type)
        if vid in llist:
            return llist[vid]
        if vid in (7,8):
            if (15-vid) in llist:
                return llist[15-vid]
        elif vid > 0:
            vn=(vid-1) % 3
            vg=(vid-1) // 3
            while vn > 0:
                vid=3*vg+vn
                if vid in llist:
                    return llist[vid]
                vn -= 1
        for vid in range(1,9):
            if vid in llist:
                return llist[vid]
        
        log("Cannot find voice %s",new_voice_type)
        return v
    
    def correct_settings(self):
        pf=self.settings.get('pause_factor',None)
        if pf is not None:
            try:
                pf=float(pf)
            except:
                pf=None
        if pf is None or pf > 1.0:
            self.settings['pause_factor'] = 1.0
        elif pf < 0.2:
            self.settings['pause_factor'] = 0.2

    
Speaker=SapiSpeaker()
Speaker.run()
